# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

"""
This module provides functions for creating CommandLine objects from YAML
and vice versa.
"""

import json
import yaml

from . import dictionary_source
from . import scheme_validator
from .str_utils import indent
from .extended_yaml_parser import ExtendedYAMLParser
from .cli import ExtendedBool


# pylint: disable=redefined-builtin


_INHERIT = ExtendedBool.INHERIT


def option_to_yaml(dictionary):
    '''Convert an option dictionary to YAML.'''

    option_strings = dictionary['option_strings']

    fields = [
        # name,             default,    include-if
        ('metavar',         None,       lambda v: v is not None),
        ('help',            None,       lambda v: v is not None),
        ('optional_arg',    False,      lambda v: v is True),
        ('groups',          None,       lambda v: v is not None),
        ('repeatable',      _INHERIT,   lambda v: v != _INHERIT),
        ('final',           False,      lambda v: v is True),
        ('hidden',          False,      lambda v: v is True),
        ('complete',        None,       lambda v: v is not None),
        ('nosort',          None,       lambda v: v is True),
        ('when',            None,       lambda v: v is not None),
        ('capture',         None,       lambda v: v is not None),
        ('long_opt_arg_sep', _INHERIT,  lambda v: v != _INHERIT),
    ]

    r = [f'- option_strings: {json.dumps(option_strings)}']

    for name, default, include_if in fields:
        value = dictionary.get(name, default)

        if include_if(value):
            r.append(f'  {name}: {json.dumps(value)}')

    return '\n'.join(r)


def positional_to_yaml(dictionary):
    '''Convert a positional dictionary to YAML.'''

    number = dictionary['number']

    fields = [
        # name,             default,    include-if
        ('metavar',         None,       lambda v: v is not None),
        ('help',            None,       lambda v: v is not None),
        ('repeatable',      False,      lambda v: v is True),
        ('complete',        None,       lambda v: v is not None),
        ('nosort',          None,       lambda v: v is True),
        ('when',            None,       lambda v: v is not None),
        ('capture',         None,       lambda v: v is not None),
    ]

    r = [f'- number: {number}']

    for name, default, include_if in fields:
        value = dictionary.get(name, default)

        if include_if(value):
            r.append(f'  {name}: {json.dumps(value)}')

    return '\n'.join(r)


def to_yaml(dictionary):
    '''Convert a single dictionary to YAML.'''

    prog                = dictionary['prog']
    aliases             = dictionary.get('aliases',             None)
    help                = dictionary.get('help',                None)
    wraps               = dictionary.get('wraps',               None)
    abbreviate_commands = dictionary.get('abbreviate_commands', _INHERIT)
    abbreviate_options  = dictionary.get('abbreviate_options',  _INHERIT)
    inherit_options     = dictionary.get('inherit_options',     _INHERIT)
    options             = dictionary.get('options',             None)
    positionals         = dictionary.get('positionals',         None)

    r = 'prog: %s\n' % json.dumps(prog)

    if aliases is not None:
        r += 'aliases: %s\n' % json.dumps(aliases)

    if help is not None:
        r += 'help: %s\n' % json.dumps(help)

    if wraps is not None:
        r += 'wraps: %s\n' % json.dumps(wraps)

    if abbreviate_commands != ExtendedBool.INHERIT:
        r += 'abbreviate_commands: %s\n' % json.dumps(abbreviate_commands)

    if abbreviate_options != ExtendedBool.INHERIT:
        r += 'abbreviate_options: %s\n' % json.dumps(abbreviate_options)

    if inherit_options != ExtendedBool.INHERIT:
        r += 'inherit_options: %s\n' % json.dumps(inherit_options)

    if options:
        r += 'options:\n'
        for option in options:
            r += indent(option_to_yaml(option), 2)
            r += '\n\n'

    if positionals:
        r += 'positionals:\n'
        for positional in positionals:
            r += indent(positional_to_yaml(positional), 2)
            r += '\n\n'

    return r.rstrip()


def commandline_to_yaml(commandline):
    '''Convert a cli.CommandLine object to YAML.'''

    dictionaries = dictionary_source.commandline_to_dictionaries(commandline)
    r = []

    for dictionary in dictionaries:
        r.append(to_yaml(dictionary))

    return '\n---\n'.join(r)


def replace_defines_in_documents(documents):
    '''Replaces defines in a list of documents.

    Example:
        prog: '%defines%'
        my_completer: ['choices', ['foo', 'bar']]
        ---
        prog: 'example'
        options:
          - option_strings: ['-o']
            complete: 'my_completer'
    '''

    defines = None

    for document in documents:
        if not isinstance(document.value, dict):
            continue

        if 'prog' in document.value and document.value['prog'].value == '%defines%':
            defines = document
            break

    if defines is None:
        return documents

    documents.remove(defines)
    defines.value.pop('prog')

    def replace_defines(obj):
        if isinstance(obj.value, str):
            return defines.value.get(obj.value, obj)
        if isinstance(obj.value, list):
            obj.value = [replace_defines(sub) for sub in obj.value]
            return obj
        if isinstance(obj.value, dict):
            obj.value = {key: replace_defines(val) for key, val in obj.value.items()}
            return obj
        return obj

    new = []
    for document in documents:
        new.append(replace_defines(document))
    return new


def load_from_file(file):
    '''Load a YAML file and turn it into a cli.CommandLine object.'''

    # First, load using normal yaml loader. It gives better error messages
    # in case of syntax errors.
    with open(file, 'r', encoding='utf-8') as fh:
        dictionaries = list(yaml.safe_load_all(fh))

    # Then load using extended yaml loader...
    with open(file, 'r', encoding='utf-8') as fh:
        content = fh.read()

    # Validate YAML...
    parser = ExtendedYAMLParser()
    parsed = parser.parse(content)
    parsed = replace_defines_in_documents(parsed)
    scheme_validator.validate(parsed)

    # Finally convert the config structure to CommandLine objects
    return dictionary_source.dictionaries_to_commandline(dictionaries)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Module for generating conditional code in Zsh.'''

from . import when
from . import shell


def _generate(query, tokens):
    '''Turn tokens/objects into condition code.'''

    r = []

    for obj in tokens:
        if isinstance(obj, when.OptionIs):
            r.append(_generate_option_is(query, obj))

        elif isinstance(obj, when.HasOption):
            r.append(_generate_has_option(query, obj))

        elif obj in ('&&', '||', '!'):
            r.append(obj)

        elif obj == '(':
            r.append('{')

        elif obj == ')':
            r.append(';}')

        else:
            raise AssertionError("Not reached")

    if when.needs_braces(tokens):
        return '{ %s; }' % ' '.join(r)

    return ' '.join(r)


def _generate_option_is(query, obj):
    func = query.use('option_is')
    args = [func, 'option_is', *obj.options, '--', *obj.values]
    return shell.join_quoted(args)


def _generate_has_option(query, obj):
    func = query.use('has_option')
    args = [func, 'has_option', *obj.options]
    return shell.join_quoted(args)


def generate_when_conditions(query, when_):
    '''Generate when condition code.'''

    tokens = when.parse_when(when_)
    return _generate(query, tokens)



# =============================================================================
# cli.py
# =============================================================================

def get_command_paths_pattern(self):
    paths = []
    for cmdline in self.get_parents(include_self=True):
        paths.append(cmdline.get_all_commands(with_aliases=True))

    return '|'.join(shell.escape(" ".join(combo)) for combo in product(*paths))

def get_all_commands(self, with_aliases=True):
    r = [self.prog]
    if with_aliases:
        r.extend(self.aliases)
    return r

# =============================================================================
# bash.py
# =============================================================================

def make_switch_case_pattern(option_strings):
    '''
    Generates a switch case pattern for `option_strings'.
    '''
    r = []
    for option_string in sorted(option_strings):
        if len(option_string) == 2:
            r.append('-!(-*)' + shell.escape(option_string[1]))
        else:
            r.append(shell.escape(option_string))

    return '|'.join(r)

# =============================================================================

    def _generate_options_parsing(self):
        def get_long_options_case_without_arg(options):
            return '|'.join(options)

        def get_long_options_case_with_arg(options):
            return '|'.join('%s=*' % o for o in options)

        def get_short_options_case(options):
            return '|'.join(o[1] for o in options)

        options = self.commandline.get_options(with_parent_options=True)

        if self.commandline.abbreviate_options:
            abbreviations = self._get_AbbreviationGeneratorForOptions(options)
        else:
            abbreviations = utils.DummyAbbreviationGenerator()

        local = ' '.join(
            '%s=0 %s=\'\'' % (
                make_option_variable_name(o, prefix='HAVE_'),
                make_option_variable_name(o, prefix='VALUE_')
            )
            for o in options
        )

        case_long_options = []
        case_short_options = []

        for option in options:
            long_options  = abbreviations.get_many_abbreviations(option.get_long_option_strings())
            long_options += abbreviations.get_many_abbreviations(option.get_old_option_strings())
            short_options = option.get_short_option_strings()

            if long_options:
                if option.takes_args == '?':
                    r  = '%s)\n'      % get_long_options_case_without_arg(long_options)
                    r += '  %s=1;;\n' % make_option_variable_name(option, prefix='HAVE_')
                    r += '%s)\n'      % get_long_options_case_with_arg(long_options)
                    r += '  %s=1\n'   % make_option_variable_name(option, prefix='HAVE_')
                    r += '  %s="${arg#*=}";;' % make_option_variable_name(option, prefix='VALUE_')
                    case_long_options.append(r)
                elif option.takes_args:
                    r  = '%s)\n'      % get_long_options_case_without_arg(long_options)
                    r += '  %s=1\n'   % make_option_variable_name(option, prefix='HAVE_')
                    r += '  %s="${words[++argi]}";;\n' % make_option_variable_name(option, prefix='VALUE_')
                    r += '%s)\n'      % get_long_options_case_with_arg(long_options)
                    r += '  %s=1\n'   % make_option_variable_name(option, prefix='HAVE_')
                    r += '  %s="${arg#*=}";;' % make_option_variable_name(option, prefix='VALUE_')
                    case_long_options.append(r)
                else:
                    r  = '%s)\n' % get_long_options_case_without_arg(long_options)
                    r += '  %s=1;;' % make_option_variable_name(option, prefix='HAVE_')
                    case_long_options.append(r)

            if short_options:
                if option.takes_args == '?':
                    r  = '%s)\n'    % get_short_options_case(short_options)
                    r += '  %s=1\n' % make_option_variable_name(option, prefix='HAVE_')
                    r += '  if $has_trailing_chars; then\n'
                    r += '    %s="${arg:$((i + 1))}"\n' % make_option_variable_name(option, 'VALUE_')
                    r += '  fi\n'
                    r += '  break;;'
                    case_short_options.append(r)
                elif option.takes_args:
                    r  = '%s)\n'    % get_short_options_case(short_options)
                    r += '  %s=1\n' % make_option_variable_name(option, prefix='HAVE_')
                    r += '  if $has_trailing_chars\n'
                    r += '  then %s="${arg:$((i + 1))}"\n' % make_option_variable_name(option, 'VALUE_')
                    r += '  else %s="${words[++argi]}"\n' % make_option_variable_name(option, 'VALUE_')
                    r += '  fi\n'
                    r += '  break;;'
                    case_short_options.append(r)
                else:
                    r  = '%s)\n'    % get_short_options_case(short_options)
                    r += '  %s=1;;' % make_option_variable_name(option, prefix='HAVE_')
                    case_short_options.append(r)

        s = '''\
local -a POSITIONALS
local END_OF_OPTIONS=0 POSITIONAL_NUM=0 %LOCALS%

local argi
for ((argi=1; argi < ${#words[@]} - 1; ++argi)); do
  local arg="${words[argi]}"

  case "$arg" in
%CASE_LONG_OPTIONS%
    --)
      END_OF_OPTIONS=1
      for ((++argi; argi < ${#words[@]}; ++argi)); do
        POSITIONALS[POSITIONAL_NUM++]="${words[argi]}"
      done
      break;;
    --*)
      ;;
    -)
      POSITIONALS[POSITIONAL_NUM++]="-";;
    -*)
      local i
      for ((i=1; i < ${#arg}; ++i)); do
        local char="${arg:$i:1}"
        local has_trailing_chars=$( (( i + 1 < ${#arg} )) && echo true || echo false)
        case "$char" in
%CASE_SHORT_OPTIONS%
        esac
      done;;
    *)
      POSITIONALS[POSITIONAL_NUM++]="$arg";;
  esac
done

for ((; argi < ${#words[@]}; ++argi)); do
  local arg="${words[$argi]}"

  case "$arg" in
    -) POSITIONALS[POSITIONAL_NUM++]="$arg";;
    -*);;
    *) POSITIONALS[POSITIONAL_NUM++]="$arg";;
  esac
done'''
        s = s.replace('%LOCALS%', local)

        if len(case_long_options):
            s = s.replace('%CASE_LONG_OPTIONS%', utils.indent('\n'.join(case_long_options), 4))
        else:
            s = s.replace('%CASE_LONG_OPTIONS%\n', '')

        if len(case_short_options):
            s = s.replace('%CASE_SHORT_OPTIONS%', utils.indent('\n'.join(case_short_options), 10))
        else:
            s = s.replace('%CASE_SHORT_OPTIONS%\n', '')

        return s

# =============================================================================

_BASH_HELPER = helpers.ShellFunction('bash_helper', r'''
# ===========================================================================
#
# This function is for querying the command line.
#
# COMMANDS
#   setup <OPTIONS> <ARGS...>
#     This is the first call you have to make, otherwise the other commands 
#     won't (successfully) work.
#
#     It parses <ARGS> accordings to <OPTIONS> and stores results in the 
#     variables POSITIONALS, HAVING_OPTIONS and OPTION_VALUES.
#
#     The first argument is a comma-seperated list of options that the parser
#     should know about. Short options (-o), long options (--option), and
#     old-style options (-option) are supported.
#
#     If an option takes an argument, it is suffixed by '='.
#     If an option takes an optional argument, it is suffixed by '=?'.
#
#   get_positional <NUM>
#     Prints out the positional argument number NUM (starting from 1)
#
#   has_option <OPTIONS...>
#     Checks if a option given in OPTIONS is passed on commandline.
#
#   option_is <OPTIONS...> -- <VALUES...>
#     Checks if one option in OPTIONS has a value of VALUES.
#
# EXAMPLE
#   local POSITIONALS HAVING_OPTIONS OPTION_VALUES
#   zsh_helper setup '-f,-a=,-optional=?' program_name -f -optional -a foo bar
#   zsh_helper has_option -f
#   zsh_helper option_is -a -- foo
#
#   Here, -f is a flag, -a takes an argument, and -optional takes an optional
#   argument.
#
#   Both queries return true.
#
# ===========================================================================

local FUNC="bash_helper"
local CONTAINS="__contains"

__contains() {
  local ARG KEY="$1"; shift
  for ARG; do [[ "$KEY" == "$ARG" ]] && return 0; done
  return 1
}

if [[ $# == 0 ]]; then
  echo "$FUNC: missing command" >&2
  return 1;
fi

local CMD=$1
shift

case "$CMD" in
  get_positional)
    if test $# -ne 1; then
      echo "$FUNC: get_positional: takes exactly one argument" >&2
      return 1;
    fi

    if test "$1" -eq 0; then
      echo "$FUNC: get_positional: positionals start at 1, not 0!" >&2
      return 1
    fi

    printf "%s" "${POSITIONALS[$1]}"
    return 0
    ;;
  has_option)
    if test $# -eq 0; then
      echo "$FUNC: has_option: arguments required" >&2
      return 1;
    fi

    local OPTION=''
    for OPTION in "${HAVING_OPTIONS[@]}"; do
      $CONTAINS "$OPTION" "$@" && return 0
    done

    return 1
    ;;
  option_is)
    local -a CMD_OPTION_IS_OPTIONS=() CMD_OPTION_IS_VALUES=()
    local END_OF_OPTIONS_NUM=0

    while test $# -ge 1; do
      if [[ "$1" == "--" ]]; then
        (( ++END_OF_OPTIONS_NUM ))
      elif test $END_OF_OPTIONS_NUM -eq 0; then
        CMD_OPTION_IS_OPTIONS+=("$1")
      elif test $END_OF_OPTIONS_NUM -eq 1; then
        CMD_OPTION_IS_VALUES+=("$1")
      fi

      shift
    done

    if test ${#CMD_OPTION_IS_OPTIONS[@]} -eq 0; then
      echo "$FUNC: option_is: missing options" >&2
      return 1
    fi

    if test ${#CMD_OPTION_IS_VALUES[@]} -eq 0; then
      echo "$FUNC: option_is: missing values" >&2
      return 1
    fi

    local I=$(( ${#HAVING_OPTIONS[@]} - 1))
    while test $I -ge 0; do
      local OPTION="${HAVING_OPTIONS[$I]}"
      if $CONTAINS "$OPTION" "${CMD_OPTION_IS_OPTIONS[@]}"; then
        local VALUE="${OPTION_VALUES[$I]}"
        $CONTAINS "$VALUE" "${CMD_OPTION_IS_VALUES[@]}" && return 0
      fi

      (( --I ))
    done

    return 1
    ;;
  setup)
    local IFS=','
    local -a OPTIONS=(${1})
    unset IFS
    shift
    ;;
  *)
    echo "$FUNC: argv[1]: invalid command" >&2
    return 1
    ;;
esac

# continuing setup....

# ===========================================================================
# Parsing of available options
# ===========================================================================

local -a   OLD_OPTS_WITH_ARG   OLD_OPTS_WITH_OPTIONAL_ARG   OLD_OPTS_WITHOUT_ARG
local -a  LONG_OPTS_WITH_ARG  LONG_OPTS_WITH_OPTIONAL_ARG  LONG_OPTS_WITHOUT_ARG
local -a SHORT_OPTS_WITH_ARG SHORT_OPTS_WITH_OPTIONAL_ARG SHORT_OPTS_WITHOUT_ARG

local OPTION
for OPTION in "${OPTIONS[@]}"; do
  case "$OPTION" in
    --?*=)    LONG_OPTS_WITH_ARG+=("${OPTION%=}");;
    --?*=\?)  LONG_OPTS_WITH_OPTIONAL_ARG+=("${OPTION%=?}");;
    --?*)     LONG_OPTS_WITHOUT_ARG+=("$OPTION");;

    -?=)      SHORT_OPTS_WITH_ARG+=("${OPTION%=}");;
    -?=\?)    SHORT_OPTS_WITH_OPTIONAL_ARG+=("${OPTION%=?}");;
    -?)       SHORT_OPTS_WITHOUT_ARG+=("$OPTION");;

    -??*=)    OLD_OPTS_WITH_ARG+=("${OPTION%=}");;
    -??*=\?)  OLD_OPTS_WITH_OPTIONAL_ARG+=("${OPTION%=?}");;
    -??*)     OLD_OPTS_WITHOUT_ARG+=("$OPTION");;

    *) echo "$FUNC: $OPTION: not a valid short, long or oldstyle option" >&2; return 1;;
  esac
done

# ===========================================================================
# Parsing of command line options
# ===========================================================================

POSITIONALS=()
HAVING_OPTIONS=()
OPTION_VALUES=()

local ARGI=2 # ARGI[1] is program name
while [[ $ARGI -le $# ]]; do
  local ARG="${!ARGI}"
  local HAVE_TRAILING_ARG=$(test $ARGI -lt $# && echo true || echo false)

  case "$ARG" in
    -)
      POSITIONALS+=(-);;
    --)
      for ARGI in $(seq $((ARGI + 1)) $#); do
        POSITIONALS+=("${@[$ARGI]}")
      done
      break;;
    --*)
      for OPTION in "${LONG_OPTS_WITH_ARG[@]}" "${LONG_OPTS_WITHOUT_ARG[@]}" "${LONG_OPTS_WITH_OPTIONAL_ARG[@]}"; do
        if [[ "$ARG" == "$OPTION="* ]]; then
          HAVING_OPTIONS+=("$OPTION")
          OPTION_VALUES+=("${ARG#$OPTION=}")
          break
        elif [[ "$ARG" == "$OPTION" ]]; then
          if $CONTAINS "$OPTION" "${LONG_OPTS_WITH_ARG[@]}"; then
            if $HAVE_TRAILING_ARG; then
              HAVING_OPTIONS+=("$OPTION")
              OPTION_VALUES+=("${@[$((ARGI + 1))]}")
              (( ARGI++ ))
            fi
          else
            HAVING_OPTIONS+=("$OPTION")
            OPTION_VALUES+=("")
          fi
          break
        fi
      done;;
    -*)
      local HAVE_MATCH=false

      for OPTION in "${OLD_OPTS_WITH_ARG[@]}" "${OLD_OPTS_WITHOUT_ARG[@]}" "${OLD_OPTS_WITH_OPTIONAL_ARG[@]}"; do
        if [[ "$ARG" == "$OPTION="* ]]; then
          HAVING_OPTIONS+=("$OPTION")
          OPTION_VALUES+=("${ARG#$OPTION=}")
          HAVE_MATCH=true
          break
        elif [[ "$ARG" == "$OPTION" ]]; then
          if $CONTAINS "$OPTION" "${OLD_OPTS_WITH_ARG[@]}"; then
            if $HAVE_TRAILING_ARG; then
              HAVING_OPTIONS+=("$OPTION")
              OPTION_VALUES+=("${@:$((ARGI + 1)):1}")
              (( ARGI++ ))
            fi
          else
            HAVING_OPTIONS+=("$OPTION")
            OPTION_VALUES+=("")
          fi

          HAVE_MATCH=true
          break
        fi
      done

      if ! $HAVE_MATCH; then
        local ARG_LENGTH=${#ARG}
        local I=1
        local IS_END=false
        while ! $IS_END && test $I -lt $ARG_LENGTH; do
          local ARG_CHAR="${ARG:$I:1}"
          local HAVE_TRAILING_CHARS=$(test $((I+1)) -lt $ARG_LENGTH && echo true || echo false)

          for OPTION in "${SHORT_OPTS_WITH_ARG[@]}" "${SHORT_OPTS_WITHOUT_ARG[@]}" "${SHORT_OPTS_WITH_OPTIONAL_ARG[@]}"; do
            local OPTION_CHAR="${OPTION:1:1}"

            if test "$ARG_CHAR" = "$OPTION_CHAR"; then
              if $CONTAINS "$OPTION" "${SHORT_OPTS_WITH_ARG[@]}"; then
                if $HAVE_TRAILING_CHARS; then
                  HAVING_OPTIONS+=("$OPTION")
                  OPTION_VALUES+=("${ARG:$((I+1))}")
                  IS_END=true
                elif $HAVE_TRAILING_ARG; then
                  HAVING_OPTIONS+=("$OPTION")
                  OPTION_VALUES+=("${@:$((ARGI + 1)):1}")
                  (( ARGI++ ))
                  IS_END=true
                fi
              elif $CONTAINS "$OPTION" "${SHORT_OPTS_WITH_OPTIONAL_ARG[@]}"; then
                HAVING_OPTIONS+=("$OPTION")

                if $HAVE_TRAILING_CHARS; then
                  IS_END=true
                  OPTION_VALUES+=("${ARG:$((I+1))}")
                else
                  OPTION_VALUES+=("")
                fi
              else
                HAVING_OPTIONS+=("$OPTION")
                OPTION_VALUES+=("")
              fi

              break
            fi
          done

          (( I++ ))
        done
      fi;;
    *)
      POSITIONALS+=("$ARG");;
  esac

  (( ARGI++ ))
done
''')

# =============================================================================

_EXPORTED_VARIABLES = helpers.ShellFunction('exported_variables', r'''
COMPREPLY=()

local var
for var in $(declare -p -x | sed 's/=.*//; s/.* //'); do
  [[ "$var" == "$cur"* ]] && COMPREPLY+=("$var")
done
''')

# =============================================================================

def get_subcommands_choices(subcommands_option):
    r = OrderedDict()
    for commandline in subcommands_option.subcommands:
        for command in get_all_command_variations(commandline):
            r[command] = commandline.help
    return r



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

"""
This module extends Python's built-in argparse library by adding several
custom methods.
"""

import argparse

from .cli import ExtendedBool

# We have to use implementation details of the argparse module...
# pylint: disable=protected-access


def dummy():
    '''Since this module only modifies the classes by argparse we provide this
    function so it can be called to silence warnings.'''


def _action_complete(self, command, *args):
    setattr(self, '_complete', (command, *args))
    return self


def _action_get_complete(self):
    return getattr(self, '_complete', None)


def _action_when(self, when):
    setattr(self, '_when', when)
    return self


def _action_get_when(self):
    return getattr(self, '_when', None)


def _action_set_repeatable(self, enable=True):
    setattr(self, '_repeatable', enable)
    return self


def _action_get_repeatable(self):
    return getattr(self, '_repeatable', ExtendedBool.INHERIT)


def _action_set_final(self, enable=True):
    setattr(self, '_final', enable)
    return self


def _action_get_final(self):
    return getattr(self, '_final', False)


def _parser_alias(self, alias):
    setattr(self, '_aliases', [alias])
    return self


def _parser_aliases(self, aliases):
    setattr(self, '_aliases', list(aliases))
    return self


def _parser_get_aliases(self):
    return getattr(self, '_aliases', [])


def _parser_remove_help(self):
    self._actions.pop(0)
    self._option_string_actions.pop('-h')
    self._option_string_actions.pop('--help')


argparse.Action.complete            = _action_complete
argparse.Action.get_complete        = _action_get_complete
argparse.Action.when                = _action_when
argparse.Action.get_when            = _action_get_when
argparse.Action.set_repeatable      = _action_set_repeatable
argparse.Action.get_repeatable      = _action_get_repeatable
argparse.Action.set_final           = _action_set_final
argparse.Action.get_final           = _action_get_final
argparse.ArgumentParser.alias       = _parser_alias
argparse.ArgumentParser.aliases     = _parser_aliases
argparse.ArgumentParser.get_aliases = _parser_get_aliases
argparse.ArgumentParser.remove_help = _parser_remove_help



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Module for parsing shell-like commands with logical operators.'''

from .string_stream import StringStream


# pylint: disable=too-few-public-methods


class Literal:
    '''Represents a literal string.'''

    def __init__(self, string):
        self.string = string

    def __str__(self):
        return self.string

    def __repr__(self):
        return f'Literal({self.string!r})'


class Command:
    '''Represents a command.'''

    def __init__(self):
        self.args = []

    def __repr__(self):
        return f'Command({self.args!r})'


class Lexer(StringStream):
    '''Lexer class.'''

    def parse(self):
        '''Split string into tokens (e.g. logical operators) and literals.'''

        tokens = []
        while True:
            token = self._parse_token()
            if token is not None:
                tokens.append(token)
            else:
                return tokens

    def _parse_token(self):
        c = self.peek()

        if c is None:
            return None

        if c in ('!', '(', ')'):
            self.advance(1)
            return c

        if c.isspace():
            self.advance(1)
            return self._parse_token()

        if c == '&':
            if not self.peek(1) == '&':
                raise ValueError('Single `&` found')
            self.advance(2)
            return '&&'

        if c == '|':
            if not self.peek(1) == '|':
                raise ValueError('Single `|` found')
            self.advance(2)
            return '||'

        return self._parse_literal()

    def _parse_literal(self):
        literal = ''

        while True:
            c = self.peek()

            if c is None:
                return Literal(literal)
            if c == '"':
                literal += self.parse_shell_double_quote(in_quotes=False)
            elif c == "'":
                literal += self.parse_shell_single_quote(in_quotes=False)
            elif c.isspace() or c in ('&', '|', '!', '(', ')'):
                return Literal(literal)
            else:
                self.advance(1)
                literal += c


def make_commands(tokens):
    '''Parse tokens.

    Input:
        [Literal("foo"), Literal("bar"), '&&' Literal("baz")]

    Output:
        [Command(["foo", "bar"]), '&&', Command(["baz"])]
    '''

    new_tokens = []
    current = Command()

    for token in tokens:
        if token in ('&&', '||', '!', '(', ')'):
            if current.args:
                new_tokens.append(current)
                current = Command()
            new_tokens.append(token)
        else:
            current.args.append(str(token))

    if current.args:
        new_tokens.append(current)

    return new_tokens


def check_syntax(tokens):
    '''Checks tokens for syntax errors.'''

    last = None
    parentheses = 0

    for token in tokens:
        if token == '(':
            parentheses += 1
            if last not in (None, '&&', '||', '!', '('):
                raise ValueError("Unexpected `(`")

        elif token == ')':
            parentheses -= 1
            if parentheses < 0 or last in (None, '(', '&&', '||', '!'):
                raise ValueError("Unexpected `)`")

        elif token in ('&&', '||'):
            if last in (None, '(', '&&', '||', '!'):
                raise ValueError(f"Unexpected `{token}`")

        elif token == '!':
            if last not in (None, '(', '&&', '||', '!'):
                raise ValueError("Unexpected `!`")

        last = token

    if parentheses > 0:
        raise ValueError("Unclosed `(`")

    if last is None:
        raise ValueError("No command found")


def parse(string):
    '''Parse a string and turn it into And/Or/Not/Command objects.'''

    lex_tokens = Lexer(string).parse()
    tokens = make_commands(lex_tokens)
    check_syntax(tokens)
    return tokens
